supported_embeddings = {
    'openai': 'OpenAIEmbed',
    'google': 'GoogleEmbed',
    'vertexai': 'VertexAIEmbed',
    'huggingface': 'HugginfaceEmbed',
    'fastembed': 'FastembedEmbed',
    'bge': 'BgeEmbed',
    'ollama': 'OllamaEmbed',
    'transformers': 'TransformersEmbed'
}
