from .abstract import AbstractBot

class BasicBot(AbstractBot):
    """Represents an BasicBot in Navigator.

        Each BasicBot has a name, a role, a goal, a backstory,
        and an optional language model (llm).
    """
    pass
